#define NEW_CODE_INTERRUPT          100

// Binary code to "trigger" the new code previously 
//   downloaded to interrupt NEW_CODE_INTERRUPT!
//
// NOTE: With x bytes of data, there are 5 bytes
//       of header, 1 byte of interrupt to use
//       (Interrupt 126 = 0x7E), 2 bytes of zero
//       at the end, and x-8 bytes of program!
UCHAR gaTriggerCodeBin[] =
#include "trig_bin.hex"

ULONG gaTriggerTimeOffset = sizeof(gaTriggerCodeBin) - 4;
ULONG gaNewCodeInterruptOffset;




// Fixups for binary code to "trigger" the new code 
//   previously downloaded to interrupt NEW_CODE_INTERRUPT!
//
// NOTE: With 44 bytes of data, there are 5 bytes
//       of header, 1 byte of interrupt to use
//       (Interrupt 126 = 0x7E), 2 bytes of zero
//       at the end, and 16 bytes of program!
UCHAR gaTriggerCodeFix[] =
#include "trig_fix.hex"

// Command to run code previously set into interrupt TRIGGER_CODE_INTERRUPT
//
// NOTE: The code set into TRIGGER_CODE_INTERRUPT waits for a second, then
//       lowers the USB data line, and finally runs the code set into
//       NEW_CODE_INTERRUPT or TRIGGER_CODE_INTERRUPT based upon our
//       context's usbPlugUnplug boolean.
//
UCHAR gaTriggerMeData[] = {
    0xB6, 0xC3, 0x01, 0x00, 0x06, TRIGGER_CODE_INTERRUPT,
    0,    0
};

ULONG gaTriggerInterruptOffset = 5;




////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// NOTES ABOUT COMMUNICATING WITH THE QT ENGINE:
//
// Communications with the QT Engine take the form of data blocks.
//   One or more data blocks can be concatenated together to form
//   complex communications.  All communications with the QT Engine
//   terminate with two zero bytes (0x00, 0x00).
//
//  <Data Block1> 0x00 0x00
//
//  <Data Block1> <Data Block2> 0x00 0x00
//
//  <Data Block1> <Data Block2> <Data Block3> 0x00 0x00
//
//  <Data Block1> <Data Block2> <Data Block3>  <Data Block4> 0x00 0x00
//
// Each data block is composed of a header of 5 bytes, and a variable
//   number of bytes in the body.  The header takes the following form:
//
// 0xC3 0xB6 <2 bytes for length of body> <data block type> <body of data>
//
// A binary signature of 0xC3B6 is required as the first two bytes of
//   a data block.
//
// The next two bytes specify the number of bytes in the body of data.
//   This number will vary depending upon the type of data.
//
////////////////////////////////////////////////////////////////////////////////
// The data block type is one byte, going from 0x00 to 0x09, and has
//   the following meaning:
//
// TYPE     DESCRIPTION
// ====     ===============================================
// 0x00     Write data into specific memory location
//
// 0x01     Write data into interrupt vector location
//
// 0x02     Replace interrupt vector with this data
//            Allows QT Engine to find memory location for data
//
// 0x03     Fixup interrupt vector code with this fixup data
//            NOTE: Fixups are 2 byte code offsets into which
//                  gets added the base of the fixed up code.
//                  A fixup is an offset to code that is
//                  assembled assuming a start of 0 that
//                  should have a real starting location
//                  added to it.
//
// 0x04     Force jump to location
//
// 0x05     Force call to location
//
// 0x06     Force interrupt call
//
// 0x07     Cause data to be written using the specified QT Engine
//            interrupt - from starting location in memory for a
//            specified number of bytes.  Allows optional QT Engine
//            memory location for redirection, unused for USB
//            interface (used for I2C, for example).
//
// 0x08     Cause data from stream to be written using the
//            specified QT Engine interrupt.
//
// 0x09     Cause data from stream to be written to configuration
//            locations specified (using AddressByte/DataWord tuplets).
//            NOTE: Addresses are specified assuming 0xC000 offset.
//                  To specify address 0xC024, use config address 0x24.
//
////////////////////////////////////////////////////////////////////////////////
//
// NOTE: Here are some miscellaneous implementation notes:
//
// (1) To download code modules, use interrupt 100 (0x64).
//
// (2) For each code module specified in the registry, we want to download
//     code (.Bin module), fix it up as needed, and then execute it.  The
//     LAST code module must be handled differently.
//
// (3) The last code module will have its code downloaded, then have that
//     code fixed up as necessary.  BUT, instead of calling interrupt 100
//     (0x64) directly, we have to load our own AoxLoad QTEngine code into
//     interrupt 126 (0x7E - the first of the interrupts reserved for a
//     debugger).
//
//     Our code will have the USB pullup line drop, wait far a bit, 
//     execute the downloaded code at interrupt 100 (0x64).
//
// (4) The LAST MODULE downloaded will have to turn on the USB pullup line
//     so the host machine can now see the "new" device.
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

UCHAR  gaNewCodeBin[] =
   #include "aox_bin.hex"

USHORT gNewCodeBinSize = sizeof(gaNewCodeBin);

UCHAR  gaNewCodeFix[] =
   #include "aox_fix.hex"

USHORT gNewCodeFixSize = sizeof(gaNewCodeFix);

NTSTATUS
WdmDownloadNewFirmwareIfNeeded( IN pAdapterContext pAC )
{
    NTSTATUS status;
    USHORT   runInterrupt = NEW_CODE_INTERRUPT;


    status = WdmDownloadFirmwareBin( pAC,
                                     NEW_CODE_INTERRUPT,
                                     gaNewCodeBin,
                                     gNewCodeBinSize,
                                     FALSE
                                   );

    status = WdmDownloadFirmwareFix( pAC,
                                     NEW_CODE_INTERRUPT,
                                     gaNewCodeFix,
                                     gNewCodeFixSize
                                   );

    gaTriggerMeData[gaTriggerInterruptOffset] = (UCHAR)NEW_CODE_INTERRUPT;

    status = WdmSendRomCommand( pAC,
                                0xFF,
                                6,
                                runInterrupt,
                                FALSE,
                                gaTriggerMeData,
                                sizeof(gaTriggerMeData),
                                NULL );

}


NTSTATUS
WdmDownloadFirmwareBin( IN pAdapterContext pAC,
                        IN CHAR            bInterrupt,
                        IN PUCHAR          pData,
                        IN USHORT          nSize,
                        IN BOOLEAN         bLoadHigh
                      )
{
    NTSTATUS status;

    status = WdmDownloadFirmware( pAC, bInterrupt, 2, pData, nSize, bLoadHigh );

    if ( !NT_SUCCESS(status) ) {
        USBENET_KdPrint( DBGLVL_MINIMUM,
                         ("FAILURE loading Bin into device (0x%0X)!\n", (ULONG)status) );
    }

    return status;
}


NTSTATUS
WdmDownloadFirmwareFix( IN pAdapterContext pAC,
                        IN CHAR            bInterrupt,
                        IN PUCHAR          pData,
                        IN USHORT          nSize
                      )
{
    NTSTATUS status;

    status = WdmDownloadFirmware( pAC, bInterrupt, 3, pData, nSize, FALSE );

    if ( !NT_SUCCESS(status) ) {
        USBENET_KdPrint( DBGLVL_MINIMUM,
                         ("FAILURE loading FIX into device (0x%0X)!\n", (ULONG)status) );
    }

    return status;
}


NTSTATUS
WdmDownloadFirmware( IN pAdapterContext pAC,
                     IN CHAR            bInterrupt,
                     IN UCHAR           bType,
                     IN PUCHAR          pData,
                     IN USHORT          nSize,
                     IN BOOLEAN         bLoadHigh
                   )
{
    NTSTATUS status;
    SHORT *sp;

    // Fill out the binary data with the correct bits!
    sp       = (SHORT*) pData;
    sp[1]    = nSize - 7;

    pData[4] = bType;
    pData[5] = bInterrupt;

    if ( bLoadHigh )
        pData[5] = (UCHAR)-bInterrupt;   //indicate load at end of memory
    
    status = WdmDownloadFirmwareData(pAC, pData, nSize, bType, bInterrupt);

    return status;
}



NTSTATUS
WdmDownloadFirmwareData(IN pAdapterContext pAC,
                        IN PUCHAR          pData,
                        IN ULONG           dwSize,
                        IN UCHAR           bType,
                        IN UCHAR           bInterrupt
                       )
{
    PURB     pUrb;
    NTSTATUS status;

    MemAlloc( pAC->pNdisContext,
              sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
              (PUCHAR *)&pUrb );

    USBENET_ASSERT( pUrb != NULL );

    if ( pUrb == NULL ) {
        USBENET_KdPrint( DBGLVL_MINIMUM,("WdmDownLoadData(): PURB alloc FAILED!\n") );
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlZeroMemory( pUrb, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST) );

    UsbBuildVendorRequest(pUrb,                         // PURB
                          URB_FUNCTION_VENDOR_DEVICE,   // Function
      sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), // size of URB
                          USBD_TRANSFER_DIRECTION_OUT,  // Transfer Flags
                          0,                            // Reserved Bits
                          0xFF,                         // Request
                          bType,                        // Value
                          bInterrupt,                   // Index
                          pData,                        // Xfer Buffer
                          NULL,                         // Xfer Buffer MDL
                          dwSize,                       // Xfer Buffer Len
                          NULL                          // Link to next URB
                         );

    status = WdmCallUSBD( pAC, pUrb );

    if ( NT_SUCCESS(status) )
        MemFree( pAC->pNdisContext, (PUCHAR)pUrb );
    else
        USBENET_ASSERT( NT_SUCCESS(status) );

    return status;
}

NTSTATUS
WdmSendRomCommand( IN pAdapterContext pAC,
                   IN UCHAR           request,
                   IN USHORT          value,
                   IN USHORT          index,
                   IN BOOLEAN         bIn,
                   IN PUCHAR          pData,
                   IN UCHAR           dataSize,
                   OUT PUCHAR         pByteCount
                 )
{
    PURB     pUrb;
    NTSTATUS status;
    
    struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST *pvr;
    
    MemAlloc( pAC->pNdisContext,
              sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST),
              (PUCHAR *)&pUrb );

    USBENET_ASSERT( pUrb != NULL );

    if ( pUrb == NULL ) {
        USBENET_KdPrint( DBGLVL_MINIMUM,("WdmSendRomCommand(): PURB alloc FAILED!\n") );
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    RtlZeroMemory( pUrb, sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST) );

    UsbBuildVendorRequest(pUrb,                         // PURB
                          URB_FUNCTION_VENDOR_DEVICE,   // Function
      sizeof(struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST), // size of URB
                          bIn                           // Transfer Flags
                          ? USBD_SHORT_TRANSFER_OK |
                          USBD_TRANSFER_DIRECTION_IN
                          : USBD_TRANSFER_DIRECTION_OUT,  
                          0,                            // Reserved Bits
                          request,                      // Request
                          value,                        // Value
                          index,                        // Index
                          pData,                        // Xfer Buffer
                          NULL,                         // Xfer Buffer MDL
                          dataSize,                     // Xfer Buffer Len
                          NULL
                         );

    // To read back data, we need to put a location in the Index...
    pvr = &pUrb->UrbControlVendorClassRequest;

    status = WdmCallUSBD( pAC, pUrb );

    if ( NT_SUCCESS(status) ) {

        if ( pByteCount )
            *pByteCount = (UCHAR)pvr->TransferBufferLength;

        MemFree( pAC->pNdisContext, (PUCHAR)pUrb );

    } else {
        USBENET_ASSERT( NT_SUCCESS(status) );

        if ( pByteCount )
            *pByteCount = 0;
    }

    return status;
}

NTSTATUS
WdmCallUSBD(
  IN pAdapterContext pAC,
  IN PURB            pUrb,
  )
/*++

Routine Description:

  Pass the USBD driver the specified URB. 


Arguments:

  pAC - pointer to the current adapter context

  pUrb - pointer to the URB to pass to the USBD driver


Comments:


Return Value:

  Status of the operation.

--*/
{
  NTSTATUS            ntStatus;
  BOOLEAN             status;
  PIRP                pIrp;
  PIO_STACK_LOCATION  nextStack;

  // Issue a synchronous request to read the USB device
  NdisInitializeEvent((PNDIS_EVENT)&gWdmEvent);
  NdisResetEvent((PNDIS_EVENT)&gWdmEvent);
  
  pIrp = IoBuildDeviceIoControlRequest(
      IOCTL_INTERNAL_USB_SUBMIT_URB,
      pAC->pPDO,
      NULL,
      0,
      NULL,
      0,
      TRUE, /* INTERNAL */
      &gWdmEvent,
      &gWdmIoStatus);
  
  // Call the class driver to perform the operation.
  
  // If the returned status is STATUS_PENDING,
  //   wait for the request to complete.
  nextStack = IoGetNextIrpStackLocation(pIrp);
  ASSERT(nextStack != NULL);
  
  // pass the URB to the USBD 'class driver'
  nextStack->Parameters.Others.Argument1 = pUrb;
  
  NdisResetEvent((PNDIS_EVENT)&gWdmEvent);

  ntStatus = IoCallDriver(pAC->pPDO, pIrp);
  
  if(ntStatus == STATUS_PENDING)
  {
    {
      status = NdisWaitEvent((PNDIS_EVENT)&gWdmEvent, TimeOut);
    }

    if(!status)
    {
      IoCancelIrp(pIrp);
      ntStatus = STATUS_IO_TIMEOUT;
      return ntStatus;
    }
  }
  else
    gWdmIoStatus.Status = ntStatus;
  
  // USBD maps the error code for us
  ntStatus = gWdmIoStatus.Status;
  
  return ntStatus;

}   // WdmCallUSBD

